// OriginalView.cpp : implementation file
//

#include "stdafx.h"
#include "Impressionist.h"
#include "ImpressionistDoc.h"
#include "OriginalView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// COriginalView

IMPLEMENT_DYNCREATE(COriginalView, CScrollView)

COriginalView::COriginalView()
{
}

COriginalView::~COriginalView()
{
}


BEGIN_MESSAGE_MAP(COriginalView, CScrollView)
	//{{AFX_MSG_MAP(COriginalView)
	ON_WM_ERASEBKGND()
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// COriginalView drawing

void COriginalView::OnInitialUpdate()
{
	CScrollView::OnInitialUpdate();

	// nothing shown in the window, so set some arbitrary size for the
	// scrollbars

	CSize sizeTotal;
	sizeTotal.cx = sizeTotal.cy = 100;
	SetScrollSizes(MM_TEXT, sizeTotal);
}

void COriginalView::OnDraw(CDC* pDC)
{
	CImpressionistDoc* pDoc = (CImpressionistDoc*) GetDocument();

	::wglMakeCurrent( m_hDC, m_hRC );

	glClear( GL_COLOR_BUFFER_BIT );

	if ( pDoc->m_ucBitmap ) {

		// note that both OpenGL pixel storage and the Windows BMP format
		// store pixels left-to-right, BOTTOM-to-TOP!!  thus all the fiddling
		// around with startrow.

		int drawWidth, drawHeight;
		GLvoid* bitstart;
		
		CPoint scrollpos = GetScrollPosition();

		drawWidth = min( windowWidth, pDoc->m_nWidth );
		drawHeight = min( windowHeight, pDoc->m_nHeight );

		int startrow = pDoc->m_nHeight - (scrollpos.y + drawHeight);
		if ( startrow < 0 ) startrow = 0;
		bitstart = pDoc->m_ucBitmap + 3 * ((pDoc->m_nWidth * startrow) + scrollpos.x);

		glRasterPos2i( 0, windowHeight - drawHeight );
		glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
		glPixelStorei( GL_UNPACK_ROW_LENGTH, pDoc->m_nWidth );
		glDrawBuffer( GL_BACK );
		glDrawPixels( drawWidth, drawHeight, GL_RGB, GL_UNSIGNED_BYTE, bitstart );

	}
			
	glFlush();

	if ( FALSE == ::SwapBuffers( m_hDC ) )	TRACE( "swapbuffers failed\n" );
	::wglMakeCurrent( NULL, NULL );
}

/////////////////////////////////////////////////////////////////////////////
// COriginalView diagnostics

#ifdef _DEBUG
void COriginalView::AssertValid() const
{
	CScrollView::AssertValid();
}

void COriginalView::Dump(CDumpContext& dc) const
{
	CScrollView::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// COriginalView message handlers

BOOL COriginalView::PreCreateWindow(CREATESTRUCT& cs) 
{
	cs.style |= WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CS_OWNDC;
	
	return CScrollView::PreCreateWindow(cs);
}

BOOL COriginalView::OnEraseBkgnd(CDC* pDC) 
{
	return TRUE;	
}

int COriginalView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CScrollView::OnCreate(lpCreateStruct) == -1)
		return -1;
		
	InitializeOpenGL();

	return 0;
}

//
// this code based on Fosner, "OpenGL Programming for Windows 95 and Windows NT".
//

BOOL COriginalView::InitializeOpenGL( void )
{
	PIXELFORMATDESCRIPTOR pfd = {
		sizeof( PIXELFORMATDESCRIPTOR ),
		1,
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		24,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 
		0, 0, 0, 0, 
		16,
		0, 0, 
		PFD_MAIN_PLANE,
		0, 0, 0, 0 };
	int pixelformat;

	m_hDC = ::GetDC( m_hWnd );

	if ( 0 == ( pixelformat = ::ChoosePixelFormat( m_hDC, &pfd ) ) ) {
		TRACE( "ChoosePixelFormat failed\n" );
		return FALSE;
	}

	if ( FALSE == ::SetPixelFormat( m_hDC, pixelformat, &pfd ) ) {
		TRACE( "SetPixelFormat failed\n" );
		return FALSE;
	}

	m_hRC = ::wglCreateContext( m_hDC );

	::wglMakeCurrent( m_hDC, m_hRC );
	InitializeGL();
	::wglMakeCurrent( NULL, NULL );

	return TRUE;
}

void COriginalView::InitializeGL( void )
{
	glClearColor( 0.4f, 0.4f, 0.4f, 0.0f );
	glDisable( GL_DEPTH_TEST );
}


void COriginalView::OnDestroy() 
{
	CScrollView::OnDestroy();
	
	::wglMakeCurrent( NULL, NULL );
	::wglDeleteContext( m_hRC );
	::ReleaseDC( m_hWnd, m_hDC );
}

void COriginalView::OnSize(UINT nType, int cx, int cy) 
{
	CScrollView::OnSize(nType, cx, cy);
	
	if ( 0 >= cx || 0 >= cy )
		return;

	// when the window is resized, we need to adjust OpenGL's
	// viewport and projection matrix.

	RECT clientRect;
	GetClientRect( &clientRect );
	cx = clientRect.right;
	cy = clientRect.bottom;

	windowWidth = cx;
	windowHeight = cy;

	::wglMakeCurrent( m_hDC, m_hRC );

	glViewport( 0, 0, cx, cy);

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluOrtho2D( 0, cx, 0, cy );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glTranslatef( 0.375, 0.375, 0.0 );

	::wglMakeCurrent( NULL, NULL );
}


void COriginalView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
{
	CImpressionistDoc* pDoc = (CImpressionistDoc*) GetDocument();

	switch ( lHint ) {

	case UPDATE_HINT_NEWORIG:

		// the user has selected a new original image.  reset the view sizes, 
		// and invalidate to force redraw.

		if ( pDoc->m_ucBitmap == NULL ) 
			SetScrollSizes( MM_TEXT, CSize(256, 256) );
		else 
			SetScrollSizes( MM_TEXT, CSize(pDoc->m_nWidth, pDoc->m_nHeight) );
		Invalidate();

		break;

	}
}

