// PaintView.cpp : implementation of the CPaintView class
//

#include "stdafx.h"
#include "Impressionist.h"

#include "ImpressionistDoc.h"
#include "PaintView.h"
#include "BrushDialog.h"

#include <math.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CPaintView

IMPLEMENT_DYNCREATE(CPaintView, CScrollView)

BEGIN_MESSAGE_MAP(CPaintView, CScrollView)
	//{{AFX_MSG_MAP(CPaintView)
	ON_WM_ERASEBKGND()
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_SIZE()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	ON_WM_MOUSEMOVE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CPaintView construction/destruction

CPaintView::CPaintView() :
	m_mtTracking(None)
{
}

CPaintView::~CPaintView()
{
}

BOOL CPaintView::PreCreateWindow(CREATESTRUCT& cs)
{
	cs.style |= WS_CLIPSIBLINGS | WS_CLIPCHILDREN;

	return CScrollView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CPaintView drawing

void CPaintView::OnDraw(CDC* pDC)
{
	CImpressionistDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	::wglMakeCurrent( m_hDC, m_hRC );

	glDrawBuffer( GL_BACK );
	glClear( GL_COLOR_BUFFER_BIT );

	if ( pDoc->m_ucPainting ) {

		CPoint scrollpos = GetScrollPosition();

		int drawWidth, drawHeight;
		drawWidth = min( m_nWindowWidth, pDoc->m_nPaintWidth );
		drawHeight = min( m_nWindowHeight, pDoc->m_nPaintHeight );

		int startrow = pDoc->m_nPaintHeight - (scrollpos.y + drawHeight);
		if ( startrow < 0 ) startrow = 0;
		m_pPaintBitstart = pDoc->m_ucPainting + 
			3 * ((pDoc->m_nPaintWidth * startrow) + scrollpos.x);

		glRasterPos2i( 0, m_nWindowHeight - drawHeight );
		glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
		glPixelStorei( GL_UNPACK_ROW_LENGTH, pDoc->m_nPaintWidth );
		glDrawPixels( drawWidth, drawHeight, GL_RGB, GL_UNSIGNED_BYTE, m_pPaintBitstart);

		m_nDrawWidth = drawWidth;
		m_nDrawHeight = drawHeight;

		m_nStartRow = startrow;
		m_nEndRow = startrow + drawHeight;
		m_nStartCol = scrollpos.x;
		m_nEndCol = m_nStartCol + drawWidth;
	}
			
	glFlush();

	if ( FALSE == ::SwapBuffers( m_hDC ) )
		TRACE( "swapbuffers failed\n" );

	::wglMakeCurrent( NULL, NULL );
}

void CPaintView::OnInitialUpdate()
{
	CImpressionistDoc* pDoc = GetDocument();

	pDoc->m_pPaintView = this;

	// the brush dialog officially belongs to this view.
	pDoc->AllocateDialog();

	CScrollView::OnInitialUpdate();
	// no painting yet; set some arbitrary size.
	CSize sizeTotal;
	sizeTotal.cx = sizeTotal.cy = 100;
	SetScrollSizes(MM_TEXT, sizeTotal);
}

/////////////////////////////////////////////////////////////////////////////
// CPaintView diagnostics

#ifdef _DEBUG
void CPaintView::AssertValid() const
{
	CScrollView::AssertValid();
}

void CPaintView::Dump(CDumpContext& dc) const
{
	CScrollView::Dump(dc);
}

CImpressionistDoc* CPaintView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CImpressionistDoc)));
	return (CImpressionistDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CPaintView message handlers

void CPaintView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
{
	CImpressionistDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	switch ( lHint )
	{
	case UPDATE_HINT_NEWORIG:

		// user has selected a new original image.  clear the current painting,
		// resize the view, and redraw.

		if ( pDoc->m_ucBitmap == NULL ) {
			SetScrollSizes( MM_TEXT, CSize(256, 256) );
		} else {
			SetScrollSizes( MM_TEXT, CSize(pDoc->m_nPaintWidth, pDoc->m_nPaintHeight) );

			if ( pDoc->m_ucPainting ) delete pDoc->m_ucPainting;
			//TRACE( "allocating %d bytes\n", pDoc->m_nPaintWidth * pDoc->m_nPaintHeight * 3 );
			pDoc->m_ucPainting = new GLubyte [pDoc->m_nPaintWidth * pDoc->m_nPaintHeight * 3];

			// using the memcpy instead of the memset will make the painting start
			// as a copy of the original instead of a black rectangle.  helpful
			// for debugging.

			memset( pDoc->m_ucPainting, 0, pDoc->m_nPaintWidth * pDoc->m_nPaintHeight * 3 );
			//memcpy( pDoc->m_ucPainting, pDoc->m_ucBitmap, pDoc->m_nPaintWidth * pDoc->m_nPaintHeight * 3 );
		}

		Invalidate();

		break;

	default:
		break;
	}
}

BOOL CPaintView::OnEraseBkgnd(CDC* pDC) 
{
	// we'll clear the buffer with OpenGL, so prevent Windows from erasing
	// the window to prevent flashing.
	return TRUE;
}

int CPaintView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CScrollView::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	InitializeOpenGL();

	return 0;
}

BOOL CPaintView::InitializeOpenGL( void )
{
	PIXELFORMATDESCRIPTOR pfd = {
		sizeof( PIXELFORMATDESCRIPTOR ),
		1,
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		24,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 
		0, 0, 0, 0, 
		16,
		0, 0, 
		PFD_MAIN_PLANE,
		0, 0, 0, 0 };
	int pixelformat;

	m_hDC = ::GetDC( m_hWnd );

	if ( 0 == ( pixelformat = ::ChoosePixelFormat( m_hDC, &pfd ) ) ) {
		TRACE( "ChoosePixelFormat failed\n" );
		return FALSE;
	}

	if ( FALSE == ::SetPixelFormat( m_hDC, pixelformat, &pfd ) ) {
		TRACE( "SetPixelFormat failed\n" );
		return FALSE;
	}

	m_hRC = ::wglCreateContext( m_hDC );

	::wglMakeCurrent( m_hDC, m_hRC );
	InitializeGL();
	::wglMakeCurrent( NULL, NULL );

	return TRUE;
}

void CPaintView::InitializeGL( void )
{
	glClearColor( 0.4f, 0.4f, 0.4f, 0.0f );
	glDisable( GL_DEPTH_TEST );
	glReadBuffer( GL_FRONT );
}

void CPaintView::OnDestroy() 
{
	CScrollView::OnDestroy();
	
	::wglMakeCurrent( NULL, NULL );
	::wglDeleteContext( m_hRC );
	::ReleaseDC( m_hWnd, m_hDC );
}

void CPaintView::OnSize(UINT nType, int cx, int cy) 
{
	CScrollView::OnSize(nType, cx, cy);
	
	if ( 0 >= cx || 0 >= cy )
		return;

	// when the window is resized, tell OpenGL about it.

	RECT clientRect;
	GetClientRect( &clientRect );
	cx = clientRect.right;
	cy = clientRect.bottom;

	m_nWindowWidth = cx;
	m_nWindowHeight = cy;

	::wglMakeCurrent( m_hDC, m_hRC );

	glViewport( 0, 0, cx, cy );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluOrtho2D( 0, cx, 0, cy );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	// fun home activity:  what is this translation for?
	glTranslatef( 0.375, 0.375, 0.0 );

	::wglMakeCurrent( NULL, NULL );
}

// the next five functions deal with user interaction on the painting
// canvas:  events with the left mouse button get sent to the 
// Brush{Begin,Move,End} methods of the current brush, events with the
// right mouse button invoke the Select{Begin,Move,End} methods.
// the function of these methods is left up to the implementor of the
// brush, but it is expected that the left button will do some painting
// with the brush, and the right will be used to set some brush parameters,
// or be ignored.

void CPaintView::OnRButtonDown(UINT nFlags, CPoint point)
{
	CImpressionistDoc* pDoc = GetDocument();

	if ( m_mtTracking != None ) return;

	::wglMakeCurrent( m_hDC, m_hRC );
		
	if ( pDoc ) {
		CPoint target( point.x, m_nWindowHeight - point.y );
		pDoc->m_pCurrentBrush->SelectBegin( target, pDoc->m_pBrushDlg );
	}

	m_mtTracking = Right;
	SetCapture();
}

void CPaintView::OnRButtonUp(UINT nFlags, CPoint point)
{
	CImpressionistDoc* pDoc = GetDocument();

	if ( m_mtTracking != Right ) return;
	
	if ( pDoc ) {
		CPoint target( point.x, m_nWindowHeight - point.y );
		pDoc->m_pCurrentBrush->SelectEnd( target, pDoc->m_pBrushDlg );
	}

	::wglMakeCurrent( NULL, NULL );

	m_mtTracking = None;
	ReleaseCapture();

	Invalidate();
}

// when the left button goes down, set up OpenGL to draw on the front 
// (visible) buffer.  when the button is released, we will transfer the
// pixels in the front buffer back into the internal copy of the painting.

void CPaintView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	CImpressionistDoc* pDoc = GetDocument();

	if ( m_mtTracking != None ) return;

	if ( pDoc->m_ucPainting == NULL ) return;

	::wglMakeCurrent( m_hDC, m_hRC );
	glDrawBuffer( GL_FRONT );

	if ( pDoc ) {
		CPoint source( point.x + m_nStartCol, m_nEndRow - point.y );
		CPoint target( point.x, m_nWindowHeight - point.y );
		pDoc->m_pCurrentBrush->BrushBegin( source, target, pDoc->m_pBrushDlg );
	}

	m_mtTracking = Left;
	SetCapture();
}


void CPaintView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	CImpressionistDoc* pDoc = GetDocument();

	if ( m_mtTracking != Left ) return;

	if ( pDoc ) {
		CPoint source( point.x + m_nStartCol, m_nEndRow - point.y );
		CPoint target( point.x, m_nWindowHeight - point.y );
		pDoc->m_pCurrentBrush->BrushEnd( source, target, pDoc->m_pBrushDlg );
	}

	glPixelStorei( GL_PACK_ALIGNMENT, 1 );
	glPixelStorei( GL_PACK_ROW_LENGTH, pDoc->m_nPaintWidth );
	
	glReadPixels( 0, m_nWindowHeight - m_nDrawHeight, 
		m_nDrawWidth, m_nDrawHeight, 
		GL_RGB, GL_UNSIGNED_BYTE, 
		m_pPaintBitstart );

	::wglMakeCurrent( NULL, NULL );

	m_mtTracking = None;
	ReleaseCapture();

	Invalidate();
}

void CPaintView::OnMouseMove(UINT nFlags, CPoint point) 
{
	CImpressionistDoc* pDoc = GetDocument();

	if ( pDoc == NULL ) return;

	CPoint source( point.x + m_nStartCol, m_nEndRow - point.y );
	CPoint target( point.x, m_nWindowHeight - point.y );

	switch ( m_mtTracking )
	{
	case None:
		break;

	case Left:
		pDoc->m_pCurrentBrush->BrushMove( source, target, pDoc->m_pBrushDlg );
		break;

	case Right:
		pDoc->m_pCurrentBrush->SelectMove( target, pDoc->m_pBrushDlg );
		break;

	}

}



